#include <LiquidCrystal.h>
#include <SPI.h>
#include <SD.h>
#include <Wire.h>    
#include "DueTimer.h"


// EEPROM: 
#define dev_adr 0x50    //Address of 24LC256 eeprom chip
unsigned int eep_adr = 20;  // Adresse im EEPROM
byte b1 = 0, b2 = 0;

// Zeiten Geschw. etc.
volatile unsigned long p1ms = 0, p1ms_start = 0, p1ms_fet = 0; // Programmlaufzeit in 0.1 ms:
volatile int lcd_aus = 0; // LCD-Ausgabe erfolgt nur ab und an
volatile float vv = 0.0; // Gescw. in (steps / 0.1ms)
volatile float tt = 0.0; // Zeit in (0.1ms)

// Drehencoder
volatile int enc_pin[3];
volatile int enc;

// SD-Karte:
const int chipSelect = 4;
byte sd_ok = 0;
byte sd_cd = 0;
int pos = 0;

// Stepper:
volatile int stp_soll[6], stp_ist[6], stp_0[6]; // Schritte der Motoren
volatile byte dir_pin[6]; // Stepper Dir pins: x, y, z, e0, e1, e2
volatile byte stp_pin[6]; // Stepper Step pins: x, y, z, e0, e1, e2
volatile byte ena_pin[6]; // Stepper enable pins: x, y, z, e0, e1, e2
volatile int motor; // Motorsteuerung

// Thermistoren:
volatile int adc_pin[6]; // Analog-Digital-Wandler Pins
volatile int adc_wert[6]; // Wandler Rohwerte
volatile float th_kohm[6]; // Widerstand in kOhm
volatile int i_kohm[6]; // Widerstand in kOhm, da sprintf nicht mit float funktioniert

// Endstops:
volatile int end_pin[6]; // 3 x min und 3 x max
volatile unsigned long ends = 0;

// FETs:
volatile byte fet_pin[6]; // Pins zum Schalten der FETs
volatile int fet; // FET-Steuerung

//Hilfsvariablen:
char zeile[21]; 
volatile int ii = 0, jj = 0, kk = 0, ll = 0, mm = 0;

// initialize the library with the numbers of the interface pins RS EN DB4 DB5 DB6 DB7
LiquidCrystal lcd(42, 43, 44, 45, 46, 47);

//*******************************************************************************************************************
void setup() {
  // LCD: 
  lcd.begin(20, 4); // Spalten, Zeilen
  lcd.setCursor(0, 0);  // Spalte, Zeile
  sprintf(zeile, "RADDS Test, Thermi.:");
  lcd.print(zeile);

  //  Analog-Digital-Wandler:
  adc_pin[5] = A5; // AUX-ADC
  adc_pin[4] = A4; // Thermistor-ADC
  adc_pin[3] = A3; // Thermistor-ADC
  adc_pin[2] = A2; // Thermistor-ADC
  adc_pin[1] = A1; // Thermistor-ADC
  adc_pin[0] = A0; // Thermistor-ADC
  for (ii = 0; ii <= 5; ii++){
    pinMode(adc_pin[ii], INPUT); 
  }

  // Drehencoder:
  enc_pin[0] = 48; // Encoder E = Knopf
  enc_pin[1] = 50; // Encoder B, im Rastpunkt am Übergang?
  enc_pin[2] = 52; // Encoder A, im Rastpunkt stabil
  for (ii = 0; ii <= 2; ii++){
    pinMode(enc_pin[ii], INPUT_PULLUP);
  }

  // Endstops:
  end_pin[0] = 28; // x_min
  end_pin[1] = 30; // y_min
  end_pin[2] = 32; // z_min
  end_pin[3] = 34; // x_max
  end_pin[4] = 36; // y_max
  end_pin[5] = 38; // z_max
  for (ii = 0; ii <= 5; ii++){
    pinMode(end_pin[ii], INPUT_PULLUP); // Nur für mechanische Endstops!
  }

  // FETs:
  fet_pin[0] = 7;  // FET 1 = Heizbett
  fet_pin[1] = 8;  // FET 2
  fet_pin[2] = 9;  // FET 2
  fet_pin[3] = 11; // FET 2
  fet_pin[4] = 12; // FET 2
  fet_pin[5] = 13; // FET 2
  for (ii = 0; ii <= 5; ii++){
    pinMode(fet_pin[ii], OUTPUT); 
    digitalWrite(fet_pin[ii],  LOW);
  }
  fet = 0; // Steuerung der FETs
  p1ms_fet = 10000; // 1 sec. nix machen (nach Start, am Ende von setup)
  
  // Schrittmotoren:
  // X-Achse:
  dir_pin[0] = 23; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[0] = 24; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[0] = 26; // Stepper enable pins: x, y, z, e0, e1, e2
  // Y-Achse:
  dir_pin[1] = 16; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[1] = 17; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[1] = 22; // Stepper enable pins: x, y, z, e0, e1, e2
  // Z-Achse:
  dir_pin[2] = 3; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[2] = 2; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[2] = 15; // Stepper enable pins: x, y, z, e0, e1, e2
  // E1-Achse:
  dir_pin[3] = A6; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[3] = A7; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[3] = A8; // Stepper enable pins: x, y, z, e0, e1, e2
  // E2-Achse:
  dir_pin[4] = A9; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[4] = A10; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[4] = A11; // Stepper enable pins: x, y, z, e0, e1, e2
  // E3-Achse:
  dir_pin[5] = 53; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[5] = 51; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[5] = 49; // Stepper enable pins: x, y, z, e0, e1, e2
  // PINs setzen:
  for (ii = 0; ii <= 5; ii++){
    pinMode(dir_pin[ii], OUTPUT); 
    pinMode(stp_pin[ii], OUTPUT); 
    pinMode(ena_pin[ii], OUTPUT); 
    digitalWrite(dir_pin[ii],  LOW);
    digitalWrite(stp_pin[ii],  LOW);
    digitalWrite(ena_pin[ii], HIGH);
    stp_soll[ii] = 0;
    stp_ist[ii] = 0;
    stp_0[ii] = 0; // Startpunkt (Integrationskonstante)
  }
  // Steuerung der Motoren
  vv = 0.32; // Geschwindigkeit / (stp / 0.1ms)
  motor = 0; 
  p1ms_start = p1ms;
  digitalWrite(ena_pin[motor], LOW); // enable motor

  // SD-Karte:
  pinMode(14,  INPUT_PULLUP); // SD detect
  lcd.setCursor(0, 3);   // Spalte, Zeile
  lcd.print("Init SD...");
  pinMode(chipSelect, OUTPUT);
  if (!SD.begin(chipSelect)) { // see if the card is present and can be initialized
    lcd.print("failed");
    sd_ok = 9;
  } else {
    lcd.print("OK");
    sd_ok = 2;
  }
  if (sd_ok == 2) {
    File dataFile = SD.open("SD_test.txt"); // open the file
    if (dataFile) {  // if the file is available, read from it:
      pos = 0;
      while (dataFile.available() && (pos < 20)) {
        zeile[pos++] = dataFile.read(); 
      }
      dataFile.close();
      zeile[pos] = 0;
      lcd.setCursor(0, 3);   // Spalte, Zeile
      lcd.print(zeile);
    } else { // if the file isn't open, pop up an error:
      lcd.setCursor(0, 3);   // Spalte, Zeile
      lcd.print("can't open file     ");
    } 
  }
  delay(4000); // Damit SD-Text lesbar ist
  
  // EEPROM:
  Wire.begin(); 
  b1 = 135; // Testbyte
  writeEEPROM(dev_adr, eep_adr, b1);
  b2 = readEEPROM(dev_adr, eep_adr);
  sprintf(zeile, "EEPROM: %3u       ", b2);
  lcd.setCursor(0, 3);   // Spalte, Zeile
  lcd.print(zeile);
  delay(2000); // Damit EEPROM-Text lesbar ist

  // Timer-Interrupt starten:
  Timer3.attachInterrupt(aufruf10kHz);
  Timer3.start(100); // Calls every 0.1ms
}

//*******************************************************************************************************************
void fets_steuern() {
  // FETs Steuern:
  if ((p1ms_fet + 10000) < p1ms) {
    digitalWrite(fet_pin[fet], HIGH);   // FET einschalten
    fet ++; // nächster FET
    if (fet > 5) {
      fet = 0;
    }
    p1ms_fet = p1ms;
    digitalWrite(fet_pin[fet], LOW);   // FET ausschalten    
  }
//  digitalWrite(fet_pin[0], HIGH);   // Heizbett zum Test immer ein
}

//*******************************************************************************************************************
void drehencoder() {
  enc = 0;
  enc = enc + 100 * digitalRead(48); 
  enc = enc + 10 * digitalRead(50); 
  enc = enc + 1 * digitalRead(52); 
}

//*******************************************************************************************************************
void endstops() {
  ends = 0;
  for (ll = 0; ll <= 5; ll++){
    ends = ends + pow(10, ll) * digitalRead(end_pin[ll]);
  }
}

//*******************************************************************************************************************
void lcd_ausgabe() {
  for (mm = 0; mm <= 4; mm++){
    i_kohm[mm] = th_kohm[mm]; // Da sprintf nicht mit float funktioniert
  }

  sprintf(zeile, "%4d%4d%4d%4d%4d", i_kohm[0], i_kohm[1], i_kohm[2], i_kohm[3], i_kohm[4]);
  lcd.setCursor(0, 1);   // Spalte, Zeile
  lcd.print(zeile);

  sprintf(zeile, "Encod. %03d ADC5 %4d", enc, adc_wert[5]);
  lcd.setCursor(0, 2);   // Spalte, Zeile
  lcd.print(zeile);

  sprintf(zeile, "Endstops: %06d cd%1d", ends, sd_cd);
  lcd.setCursor(0, 3);   // Spalte, Zeile
  lcd.print(zeile);
}

//*******************************************************************************************************************
void adc_auslesen() {
  for (kk = 0; kk <= 5; kk++){
    adc_wert[kk] = analogRead(adc_pin[kk]); // ersten Wert verwerfen
    adc_wert[kk] = analogRead(adc_pin[kk]);
    th_kohm[kk] = 4.7 * adc_wert[kk] / ((float) (0.0001 + 1023 - adc_wert[kk]));
    if (th_kohm[kk] > 9999) th_kohm[kk] = 9999;
  }
}

//*******************************************************************************************************************
inline void steps_berechnen() {
  // Motoren jeweils eine Umdrehung vor und wieder zurück
  tt = (float) (p1ms - p1ms_start);
  stp_soll[motor] = stp_0[motor] + (int) (vv * tt);

  if (((vv > 0.0) && (stp_soll[motor] > 3200)) || ((vv < 0.0) && (stp_soll[motor] < 0))) { // Vorwärtslauf und Rückwärtslauf steuern
    digitalWrite(ena_pin[motor], HIGH); // disable motor
    stp_0[motor] = stp_ist[motor];  // Startwert für nächste Bewegung
    motor ++;  // nächster Motor
    if (motor > 5) {
      motor = 0;  
      vv = -1.0 * vv; // Richtungsumkehr
    }
    digitalWrite(ena_pin[motor], LOW); // enable motor
    p1ms_start = p1ms;  // Startzeit für aktuellen Motor
  }
}  

//*******************************************************************************************************************
inline void steps_ausfuehren() {
  // Achsen evtl. verfahren:
  for (jj = 0; jj <= 5; jj++){
    if (stp_soll[jj] != stp_ist[jj]) {
      if (stp_soll[jj] > stp_ist[jj]) { // positive Richtung
        digitalWrite(dir_pin[jj], LOW);   // DIR
        stp_ist[jj] ++;
      } else { // negative Richtung
        digitalWrite(dir_pin[jj], HIGH);   // DIR
        stp_ist[jj] --;
      }
      digitalWrite(stp_pin[jj], HIGH);   // Step machen
      delayMicroseconds(1);
      digitalWrite(stp_pin[jj], LOW);   // Step abschliessen
    }
  }
}

//*******************************************************************************************************************
void writeEEPROM(int deviceaddress, unsigned int eeaddress, byte data) {
  Wire.beginTransmission(deviceaddress);
  Wire.write((int)(eeaddress >> 8));   // MSB
  Wire.write((int)(eeaddress & 0xFF)); // LSB
  Wire.write(data);
  Wire.endTransmission();
 
  delay(6); // EEPROM braucht etwas Zeit zum Speichern (max 5 ms)
}
 
//*******************************************************************************************************************
byte readEEPROM(int deviceaddress, unsigned int eeaddress) {
  byte rdata = 0xFF;
 
  Wire.beginTransmission(deviceaddress);
  Wire.write((int)(eeaddress >> 8));   // MSB
  Wire.write((int)(eeaddress & 0xFF)); // LSB
  Wire.endTransmission();
  Wire.requestFrom(deviceaddress, 1);
  if (Wire.available()) {
    rdata = Wire.read();
  }
  return rdata;
}

//*******************************************************************************************************************
void aufruf10kHz(){ // Interrupt für die Motorbewegung
  p1ms++;
  lcd_aus++;
  steps_berechnen();
  steps_ausfuehren();
}

//*******************************************************************************************************************
void loop() {
  adc_auslesen();
  drehencoder();
  fets_steuern();
  endstops();
  sd_cd = digitalRead(14); 
  if (lcd_aus > 100) {
    lcd_aus = 0;
    lcd_ausgabe();
  }
}






















